--
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity SDRAMCNT is
	port (
		DRAM_DI		: in  std_logic_vector(15 downto 0);	-- SDRAM data bus
		ADDRESS		: in  std_logic_vector(19 downto 0);
		INDATA		: in  std_logic_vector(7 downto 0);
		RDN			: in  std_logic;
		WRN			: in  std_logic;
		INIT		: in  std_logic;
		MEMNOINIT	: in  std_logic;
		ERRRST		: in  std_logic;
		CLK100M		: in  std_logic;
		CLK100M_DI	: in  std_logic;
		RSTN		: in  std_logic;
		DRAM_DO		: out std_logic_vector(15 downto 0);	-- SDRAM data bus
		DRAM_DOENB	: out std_logic;						-- SDRAM data output enable
		DRAM_A		: out std_logic_vector(12 downto 0);	-- SDRAM address bus (A12:N.C.)
		DRAM_CLK	: out std_logic;						-- SDRAM clock output
		DRAM_CKE	: out std_logic;						-- SDRAM clock enable
		DRAM_LDQM	: out std_logic;						-- SDRAM LowerByte Data Mask
		DRAM_UDQM	: out std_logic;						-- SDRAM UpperByte Data Mask
		DRAM_WE_N	: out std_logic;						-- SDRAM write Enable
		DRAM_CAS_N	: out std_logic;						-- SDRAM CAS
		DRAM_RAS_N	: out std_logic;						-- SDRAM RAS
		DRAM_CS_N	: out std_logic;						-- SDRAM chip select
		DRAM_BA_1	: out std_logic;						-- SDRAM Bank #1
		DRAM_BA_0	: out std_logic;						-- SDRAM Bank #0
		READY		: out std_logic;
		OUTDATA		: out std_logic_vector(7 downto 0);
		INITDONE	: out std_logic;
		INITERR		: out std_logic;
		MEMERR		: out std_logic
	);
end SDRAMCNT;

architecture RTL of SDRAMCNT is

	signal rdn_f1	: std_logic;
	signal rdn_f2	: std_logic;
	signal wrn_f1	: std_logic;
	signal wrn_f2	: std_logic;
	signal rdflag	: std_logic;

	signal init_f1	: std_logic;
	signal init_f2	: std_logic;
	signal init_f3	: std_logic;

	signal initst		: std_logic_vector(2 downto 0);
	signal initcnt		: std_logic_vector(14 downto 0);
	signal initaccnt	: std_logic_vector(6 downto 0);

	signal addcnt	: std_logic_vector(19 downto 0);

	signal accnt	: std_logic_vector(8 downto 0);

	signal domsk	: std_logic;
	signal wrdt		: std_logic_vector(14 downto 0);
	signal initdt	: std_logic_vector(14 downto 0);
	signal initdtr	: std_logic_vector(14 downto 0);
	signal initpty	: std_logic;
	signal initinv	: std_logic_vector(14 downto 0);
	signal datapty	: std_logic;
	signal wrdt_f1		: std_logic_vector(14 downto 0);
	signal datapty_f1	: std_logic;

	signal a_i		: std_logic_vector(13 downto 0);
	signal cs_n		: std_logic;
	signal ras_n	: std_logic;
	signal cas_n	: std_logic;
	signal we_n		: std_logic;
	signal cke		: std_logic;
	signal ldqm		: std_logic;
	signal udqm		: std_logic;

	signal cscnt	: std_logic_vector(2 downto 0);

	signal doenb	: std_logic;
	signal add		: std_logic_vector(19 downto 0);
	signal q_i		: std_logic_vector(15 downto 0);
	signal q_i_f1	: std_logic_vector(15 downto 0);
	signal outdata_i: std_logic_vector(14 downto 0);
	signal do_i		: std_logic_vector(15 downto 0);
	signal q_i_pty	: std_logic;
	signal dinenb	: std_logic;
	signal dinenb_f1: std_logic;

	signal ready_i		: std_logic;
	signal initerr_i	: std_logic;
	signal memerr_i		: std_logic;

	signal initdone_i	: std_logic;


begin

-- read / write pulse latch
	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			rdn_f2 <= '1';
		elsif (CLK100M'event and CLK100M = '1') then
			rdn_f2 <= rdn_f1;
		end if;
	end process;

	rdn_f1 <= RDN;


	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			wrn_f2 <= '1';
		elsif (CLK100M'event and CLK100M = '1') then
			wrn_f2 <= wrn_f1;
		end if;
	end process;

	wrn_f1 <= WRN;


	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			rdflag <= '0';
		elsif (CLK100M'event and CLK100M = '1') then
			if    (initst = "011" or initst = "101") then
				rdflag <= '0';
			elsif (initst = "100" or initst = "110") then
				rdflag <= '1';
			elsif (wrn_f1 = '0' and wrn_f2 = '1') then
				rdflag <= '0';
			elsif (rdn_f1 = '0' and rdn_f2 = '1') then
				rdflag <= '1';
			end if;
		end if;
	end process;


-- Initialize
	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			init_f1 <= '0';
			init_f2 <= '0';
			init_f3 <= '0';
		elsif (CLK100M'event and CLK100M = '1') then
			init_f1 <= INIT;
			init_f2 <= init_f1;
			init_f3 <= init_f2;
		end if;
	end process;

-- Initialize state machine
	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			initst    <= "000";
			initcnt   <= (others => '0');
			initaccnt <= (others => '0');
			addcnt    <= (others => '0');
		elsif (CLK100M'event and CLK100M = '1') then
			if (init_f2 = '1' and init_f3 = '0') then
				initst    <= "001";
				initcnt   <= (others => '0');
				initaccnt <= (others => '0');
			else
				case initst is
					when "001" =>							-- wait 200us
						if (initcnt = 20000) then
							initst    <= "010";
							initaccnt <= (others => '0');
							initcnt   <= (others => '0');
						else
							initcnt   <= initcnt + 1;
						end if;

					when "010" =>							-- sdram initialize
						if (initaccnt = 100) then
							if (MEMNOINIT = '1') then
								initst    <= "111";
								addcnt    <= (others => '0');
							else
								initst    <= "011";
								addcnt    <= (others => '0');
							end if;
							initaccnt <= (others => '0');
						else
							initaccnt <= initaccnt + 1;
						end if;

					when "011" =>							-- test data write
						if (accnt = 21) then
							if (addcnt = X"FFFFF") then
								initst    <= "100";
								addcnt    <= (others => '0');
							else
								addcnt    <= addcnt + 1;
							end if;
						end if;

					when "100" =>							-- test data read
						if (accnt = 21) then
							if (addcnt = X"FFFFF") then
								initst    <= "101";
								addcnt    <= (others => '0');
							else
								addcnt    <= addcnt + 1;
							end if;
						end if;

					when "101" =>							-- test data write
						if (accnt = 21) then
							if (addcnt = X"FFFFF") then
								initst    <= "110";
								addcnt    <= (others => '0');
							else
								addcnt    <= addcnt + 1;
							end if;
						end if;

					when "110" =>							-- test data read
						if (accnt = 21) then
							if (addcnt = X"FFFFF") then
								initst    <= "111";
								addcnt    <= (others => '0');
							else
								addcnt    <= addcnt + 1;
							end if;
						end if;


					when "111" =>
						initcnt   <= (others => '0');
						initaccnt <= (others => '0');

					when others =>
						initst    <= "000";
						initcnt   <= (others => '0');
						initaccnt <= (others => '0');
				end case;
			end if;
		end if;
	end process;


-- access counter
	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			accnt <= (others => '1');
		elsif (CLK100M'event and CLK100M = '1') then
			if (accnt = 511) then
				if (initst = "011" or initst = "100" or initst = "101" or initst = "110") then
					accnt <= (others => '0');
				elsif (wrn_f1 = '0' and wrn_f2 = '1') then
					accnt <= (others => '0');
				elsif (rdn_f1 = '0' and rdn_f2 = '1') then
					accnt <= (others => '0');
				end if;
			elsif (accnt = 21) then
				accnt <= (others => '1');
			else
				accnt <= accnt + 1;
			end if;
		end if;
	end process;

	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			ready_i <= '0';
		elsif (CLK100M'event and CLK100M = '1') then
			if (accnt = 511 and initst = "111") then
				ready_i <= '1';
			else
				ready_i <= '0';
			end if;
		end if;
	end process;

	wrdt <= "0000000" & INDATA;

	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			domsk <= '0';
		elsif (CLK100M'event and CLK100M = '1') then
			if    (accnt = 9) then
				domsk <= (not rdflag);
			elsif (accnt = 10) then
				domsk <= '0';
			end if;
		end if;
	end process;

	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			add  <= (others => '0');
			do_i <= (others => '0');
		elsif (CLK100M'event and CLK100M = '1') then
			if (initst = "001") then
				add  <= (others => '0');
				do_i <= (others => '0');
			elsif (initst = "011" or initst = "100" or initst = "101" or initst = "110") then
				add <= addcnt;
				if (domsk = '0') then
					do_i <= (others => '0');
				else
					do_i(14 downto 0) <= initdt;
					do_i(15)          <= initpty;
				end if;
			else
				if (accnt = 7) then
					add <= ADDRESS;
					wrdt_f1    <= wrdt;
					datapty_f1 <= datapty;
				end if;
				if (domsk = '0') then
					do_i <= (others => '0');
				else
					do_i(14 downto 0) <= wrdt_f1;
					do_i(15)          <= datapty_f1;
				end if;
			end if;
		end if;
	end process;

	initdt <=
		initinv xor
		"010101010101010" xor
		("0000000000" & addcnt(19 downto 15)) xor
		addcnt(14 downto 0);

	initdtr <= initdt;

	initpty <= 
		initdt( 0) xor initdt( 1) xor initdt( 2) xor initdt( 3) xor initdt( 4) xor
		initdt( 5) xor initdt( 6) xor initdt( 7) xor initdt( 8) xor initdt( 9) xor
		initdt(10) xor initdt(11) xor initdt(12) xor initdt(13) xor initdt(14) xor
		'1';

	datapty <= 
		wrdt( 0) xor wrdt( 1) xor wrdt( 2) xor wrdt( 3) xor wrdt( 4) xor
		wrdt( 5) xor wrdt( 6) xor wrdt( 7) xor wrdt( 8) xor wrdt( 9) xor
		wrdt(10) xor wrdt(11) xor wrdt(12) xor wrdt(13) xor wrdt(14) xor
		'1';

	initinv <= ("000" & X"000") when (initst = "011" or initst = "100") else ("111" & X"FFF");

	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			a_i   <= (others => '0');
			cs_n  <= '1';
			ras_n <= '1';
			cas_n <= '1';
			we_n  <= '1';
			cscnt <= "000";
		elsif (CLK100M'event and CLK100M = '1') then
			if (initst = "000") then							-- no access
				a_i   <= (others => '0');
				cs_n  <= '1';
				ras_n <= '1';
				cas_n <= '1';
				we_n  <= '1';
				cscnt <= "000";
			elsif (initst = "001") then							-- wait 200us
				a_i   <= (others => '0');						-- NOP
				cs_n  <= '0';
				ras_n <= '1';
				cas_n <= '1';
				we_n  <= '1';
				cscnt <= "000";
			elsif (initst = "010") then							-- sdram initialize

--  1   :PREA (PALL)
--  2- 4:NOP
--  5   :MRS
--  6- 8:NOP
--  9-79:CBR  (REF)
-- 80-91:NOP
-- 92   :PREA (PALL)
-- 93-99:NOP

				if (initaccnt = 1 or initaccnt = 92) then		-- PREA
					a_i   <= "00" & "0100" & "00000000";
					cs_n  <= '0';
					ras_n <= '0';
					cas_n <= '1';
					we_n  <= '0';
					cscnt <= "000";
				elsif (initaccnt = 5) then						-- MRS
					a_i   <= "00" & "0000" & "00110000";		-- CL=3 / burst / BL=1
					cs_n  <= '0';
					ras_n <= '0';
					cas_n <= '0';
					we_n  <= '0';
					cscnt <= "000";
				elsif (initaccnt = 9) then						-- CBR
					a_i   <= (others => '0');
					cs_n  <= '0';
					ras_n <= '0';
					cas_n <= '0';
					we_n  <= '1';
					cscnt <= "000";
				elsif (initaccnt =  2 or initaccnt = 6 or initaccnt = 80 or initaccnt = 93) then	-- NOP
					a_i   <= (others => '0');
					cs_n  <= '1';
					ras_n <= '1';
					cas_n <= '1';
					we_n  <= '1';
					cscnt <= "000";
				else
					if (cscnt = "111") then
						cs_n  <= '0';
					else
						cs_n  <= '1';
					end if;
					cscnt <= cscnt + 1;
				end if;

			else

--   0-  7:NOP
--   8    :ACT
--   9    :NOP
--  10    :WR / RD
--  11- 14:NOP
--  15    :PREA (PALL)
--  16- 18:NOP
--  19    :CBR  (REF)

				if    (accnt = 8) then						-- ACT
					a_i   <= "00" & add(19 downto 8);
					cs_n  <= '0';
					ras_n <= '0';
					cas_n <= '1';
					we_n  <= '1';
					cscnt <= "000";
				elsif (accnt = 10) then						-- RD / WR
					a_i   <= "00" & "0000" & add(7 downto 0);
					cs_n  <= '0';
					ras_n <= '1';
					cas_n <= '0';
					we_n  <= rdflag;
					cscnt <= "000";
				elsif (accnt = 15) then						-- PREA
					a_i   <= "00" & "0100" & "00000000";
					cs_n  <= '0';
					ras_n <= '0';
					cas_n <= '1';
					we_n  <= '0';
					cscnt <= "000";
				elsif (accnt = 19) then						-- CBR
					a_i   <= (others => '0');
					cs_n  <= '0';
					ras_n <= '0';
					cas_n <= '0';
					we_n  <= '1';
					cscnt <= "000";
				elsif (accnt = 0 or accnt = 9 or accnt = 11 or accnt = 16) then		-- NOP
					a_i   <= (others => '0');
					cs_n  <= '1';
					ras_n <= '1';
					cas_n <= '1';
					we_n  <= '1';
					cscnt <= "000";
				else
					if (cscnt = "111") then
						cs_n  <= '0';
					else
						cs_n  <= '1';
					end if;
					cscnt <= cscnt + 1;
				end if;

			end if;
		end if;
	end process;

	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			doenb <= '0';
		elsif (CLK100M'event and CLK100M = '1') then
			if    (accnt = 9) then
				doenb <= (not rdflag);
			elsif (accnt = 12) then
				doenb <= '0';
			end if;
		end if;
	end process;


	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			initdone_i  <= '0';
		elsif (CLK100M'event and CLK100M = '1') then
			if (initst = "111") then
				initdone_i  <= '1';
			end if;
		end if;
	end process;

	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			cke  <= '0';
			ldqm <= '1';
			udqm <= '1';
		elsif (CLK100M'event and CLK100M = '1') then
			if (initst = "001") then
				cke  <= '1';
				ldqm <= '0';
				udqm <= '0';
			end if;
		end if;
	end process;


-- output data
	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			DRAM_DO    <= (others => '0');
			DRAM_DOENB <= '0';
			DRAM_A     <= (others => '0');
			DRAM_BA_0  <= '0';
			DRAM_BA_1  <= '0';
			DRAM_CS_N  <= '1';
			DRAM_RAS_N <= '1';
			DRAM_CAS_N <= '1';
			DRAM_WE_N  <= '1';
			DRAM_CKE   <= '0';
			DRAM_LDQM  <= '1';
			DRAM_UDQM  <= '1';
		elsif (CLK100M'event and CLK100M = '0') then
			DRAM_DO    <= do_i;
			DRAM_DOENB <= doenb;
			DRAM_A     <= "0" & a_i(11 downto 0);
			DRAM_BA_0  <= a_i(12);
			DRAM_BA_1  <= a_i(13);
			DRAM_CS_N  <= cs_n;
			DRAM_RAS_N <= ras_n;
			DRAM_CAS_N <= cas_n;
			DRAM_WE_N  <= we_n;
			DRAM_CKE   <= cke;
			DRAM_LDQM  <= ldqm;
			DRAM_UDQM  <= udqm;
		end if;
	end process;

-- input data latch
	process (CLK100M_DI,RSTN)
	begin
		if (RSTN = '0') then
			q_i <= (others => '0');
		elsif (CLK100M_DI'event and CLK100M_DI = '1') then
			q_i <= DRAM_DI;
		end if;
	end process;

	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			q_i_f1    <= X"8000";
			q_i_pty   <= '1';
		elsif (CLK100M'event and CLK100M = '1') then
			if (dinenb = '1') then
				q_i_f1 <= q_i;
			end if;

			q_i_pty <= 
				q_i_f1( 0) xor q_i_f1( 1) xor q_i_f1( 2) xor q_i_f1( 3) xor q_i_f1( 4) xor
				q_i_f1( 5) xor q_i_f1( 6) xor q_i_f1( 7) xor q_i_f1( 8) xor q_i_f1( 9) xor
				q_i_f1(10) xor q_i_f1(11) xor q_i_f1(12) xor q_i_f1(13) xor q_i_f1(14) xor
				q_i_f1(15);

		end if;
	end process;

	outdata_i <= q_i_f1(14 downto 0);

	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			dinenb    <= '0';
			dinenb_f1 <= '0';
		elsif (CLK100M'event and CLK100M = '1') then
			if    (accnt = 14) then
				dinenb <= rdflag;
			elsif (accnt = 15) then
				dinenb <= '0';
			end if;
			dinenb_f1 <= dinenb;
		end if;
	end process;

	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			memerr_i <= '0';
		elsif (CLK100M'event and CLK100M = '1') then
			if (ERRRST = '1') then
				memerr_i <= '0';
			elsif (q_i_pty = '0') then
				memerr_i <= '1';
			end if;
		end if;
	end process;

	process (CLK100M,RSTN)
	begin
		if (RSTN = '0') then
			initerr_i <= '0';
		elsif (CLK100M'event and CLK100M = '1') then
			if (ERRRST = '1') then
				initerr_i <= '0';
			elsif (initst = "100" or initst = "110") then
				if ( dinenb_f1 = '1' and initdtr /= q_i_f1(14 downto 0) ) then
					initerr_i <= '1';
				end if;
			end if;
		end if;
	end process;

	DRAM_CLK  <= CLK100M;
	READY     <= ready_i;
	INITDONE  <= initdone_i;
	INITERR   <= initerr_i;
	MEMERR    <= memerr_i;
	OUTDATA   <= outdata_i(7 downto 0);

end RTL;
